#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text, List

import apache_beam as beam

from zenml.steps.inferrer import BaseInferrer


class FileWriterInferrer(BaseInferrer):
    """
    File Writer Inferrer step, writing inference results to a text file.
    """

    def __init__(self,
                 labels: List[Text],
                 output_path: Text,
                 output_ext: Text = ".csv",
                 num_shards: int = 0,
                 **kwargs):
        """
        Base Tensorflow Inferrer constructor.

        Args:
            name: Outward-facing name of the pipeline.
        """
        self.labels = labels
        self.output_path = output_path
        self.output_ext = output_ext
        self.num_shards = num_shards

        super(FileWriterInferrer, self).__init__(
            labels=labels,
            **kwargs,
        )

    def get_destination(self):
        return beam.io.WriteToText(
            self.output_path,
            file_name_suffix=self.output_ext,
            append_trailing_newlines=True,
            num_shards=self.num_shards,
        )
